// Title: Julia Quaternions Resurrection
// Based on: "Xmas Demo 2020 Edition"
// Original Author: Corneliusen
// License: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication license
// This modified version was used in the Xmas Demo 2021 by Julin & Corneliusen: https://www.ignorantus.com/pages/xmasdemo2021/
// (Loosely based on Keenan Crane's Quaternions, but the source has diverged significantly.)

#version 310 es
precision highp float;

//#define XAVIER

in vec2 v_UV1;
uniform float u_Dissolve;

#define v_texCoord v_UV1

layout(location = 0) out vec4 outColor;

layout(std140, binding = 0) uniform ctx_stuff {
    unsigned int width;
    unsigned int height;
    float aspect;
    float time;
    unsigned int frame;
};

layout(std140, binding = 2) uniform stuff
{
    vec4 start_pos;
    vec4 mu_pos;
    vec4 obj_col;
    float rot_xz;
    float ax;
    float ay;
	float starttime;
};

#ifdef XAVIER
#define ITERATIONS_NORM 10
#define ITERATIONS_INTER 10
#define BOUNDING_RADIUS_2 3.0f
#define ESCAPE_THRESHOLD 4.0f
#define DEL 0.0001
#define EPSILON              0.00750000f
#else
#define ITERATIONS_NORM 10
#define ITERATIONS_INTER 60
#define BOUNDING_RADIUS_2 3.0f
#define ESCAPE_THRESHOLD 10.0f
#define DEL 0.0001
#define EPSILON 0.0001
#endif


// Q1 Q2 = [ x1 x2 - y1 y2 - z1 z2 - w1 w2 ] + | q1.x*q2.x     - dot( q1.yzw, q2.yzw )
//         [ x1 y2 + y1 x2 + z1 w2 - w1 z2 ] + | q1.x * q2.yzw + q2x * q1.yzw + cross( q1.yzw, q2.yzw )
//         [ x1 z2 + z1 x2 + w1 y2 - y1 w2 ] + | ...             ...            ...
//         [ x1 w2 + w1 x2 + y1 z2 - z1 y2 ] + | ...             ...            ...

/*
vec4 quatMult( vec4 q1, vec4 q2 )
{
    vec4 r;
    r.x   = q1.x * q2.x - dot( q1.yzw, q2.yzw );
    r.yzw = q1.x * q2.yzw + q2.x * q1.yzw + cross( q1.yzw, q2.yzw );
    return r;
}
*/

// don't use cross() and don't replace with dot():
vec4 quatMult( vec4 q1, vec4 q2 )
{
    vec4 r;
    r.x   = q1.x * q2.x - dot( q1.yzw, q2.yzw );
    r.yzw = q1.x * q2.yzw + q2.x * q1.yzw  + q1.zwy * q2.wyz - q1.wyz * q2.zwy;
    return r;
}

vec4 quatSq( vec4 q )
{
    vec4 r;
    r.x   = q.x * q.x - dot( q.yzw, q.yzw );
    r.yzw = 2.0f * q.x * q.yzw;
    return r;
}

vec4 iterateIntersect( inout vec4 q1, vec4 c )
{
    bool r0,  r1;
    vec4 q0;
    vec4 qp0, qp1;

    // reduce for qp1=vec4( 1.0f, 0.0f, 0.0f, 0.0f )
    qp0 = 2.0f *           q1;        q0 = quatSq( q1 ) + c;
    qp1 = 2.0f * quatMult( q0, qp0 ); q1 = quatSq( q0 ) + c;
    r0  = dot( q0, q0 ) > ESCAPE_THRESHOLD;
    r1  = dot( q1, q1 ) > ESCAPE_THRESHOLD;
    if( r0 || r1 ) {
        q1   = r0 ? q0  : q1;
        return r0 ? qp0 : qp1;
    }

    // unroll 2 to reduce stalls
    for( int i = 0; i < ITERATIONS_INTER-2; i += 2 ) {
        qp0 = 2.0f * quatMult( q1, qp1 ); q0 = quatSq( q1 ) + c;
        qp1 = 2.0f * quatMult( q0, qp0 ); q1 = quatSq( q0 ) + c;
        r0  = dot( q0, q0 ) > ESCAPE_THRESHOLD;
        r1  = dot( q1, q1 ) > ESCAPE_THRESHOLD;
        if( r0 || r1 ) break;
    }

    q1   = r0 ? q0  : q1;
    return r0 ? qp0 : qp1;
}

float intersectQJulia( inout vec3 rO, vec3 rD, vec4 c )
{
    float dist;

    while( true ) {
        vec4 z  = vec4( rO, 0.0f );
        vec4 zp = iterateIntersect( z, c );

        float normZ = length( z );
        dist = 0.5f * normZ * log( normZ ) / length( zp );

        rO += rD * dist;

        if( dist < EPSILON || dot( rO, rO ) > BOUNDING_RADIUS_2 ) break;
    }

    return dist;
}

vec3 normEstimate( vec3 qP, vec4 c )
{
    vec4 gx1 = vec4( qP, 0.0f ); gx1.x -= DEL;
    vec4 gx2 = vec4( qP, 0.0f ); gx2.x += DEL;
    vec4 gy1 = vec4( qP, 0.0f ); gy1.y -= DEL;
    vec4 gy2 = vec4( qP, 0.0f ); gy2.y += DEL;
    vec4 gz1 = vec4( qP, 0.0f ); gz1.z -= DEL;
    vec4 gz2 = vec4( qP, 0.0f ); gz2.z += DEL;

    // reduce iteration 0 where z=0
    float dyz = dot( qP.yz, qP.yz );
    float qx  = qP.x * qP.x;
    float qx2 = 2.0f * qP.x;

    gx1 = vec4( gx1.x * gx1.x - dyz,                   2.0f * gx1.x * qP.yz,  0.0f ) + c;
    gx2 = vec4( gx2.x * gx2.x - dyz,                   2.0f * gx2.x * qP.yz,  0.0f ) + c;
    gy1 = vec4( qx            - dot( gy1.yz, gy1.yz ), qx2          * gy1.yz, 0.0f ) + c;
    gy2 = vec4( qx            - dot( gy2.yz, gy2.yz ), qx2          * gy2.yz, 0.0f ) + c;
    gz1 = vec4( qx            - dot( gz1.yz, gz1.yz ), qx2          * gz1.yz, 0.0f ) + c;
    gz2 = vec4( qx            - dot( gz2.yz, gz2.yz ), qx2          * gz2.yz, 0.0f ) + c;

    for( int i = 0; i < ITERATIONS_NORM-1; i++ ) {
        gx1 = quatSq( gx1 ) + c;
        gx2 = quatSq( gx2 ) + c;
        gy1 = quatSq( gy1 ) + c;
        gy2 = quatSq( gy2 ) + c;
        gz1 = quatSq( gz1 ) + c;
        gz2 = quatSq( gz2 ) + c;
    }

    return normalize( vec3( length( gx2 ) - length( gx1 ),
                            length( gy2 ) - length( gy1 ),
                            length( gz2 ) - length( gz1 ) ) );
}

vec3 Phong( vec3 light, vec3 eye, vec3 pt, vec3 N )
{
    vec3  L  = normalize( light - pt );
    vec3  E  = normalize( eye   - pt );
    float NL = dot( N, L );
    vec3  R  = L - 2.0f * NL * N;

    float diffuse = max( NL, 0.0f );

    return obj_col.xyz * diffuse + pow( max( dot( E, R ), 0.0f ), 8.0f );
}


#define PI 3.14159265f

vec3 rotate_xz( float frac, float radius )
{
    return vec3(  radius*sin( frac ),
                  0.0f,
                 -radius*cos( frac ) );
}

#ifdef XAVIER
vec4 getBG( void )
{
    vec2 df = vec2( 16.0f*3.0f, 9.0f*3.0f );
    vec3 plane_col = obj_col.yzx;
    vec2 pulse_pos = vec2( 0.0f );

    vec2 dxy = v_texCoord - 0.5f;
    dxy = floor( dxy * df )/df - pulse_pos/32.0f;
    float dv = 1.0f - sqrt( dot( dxy, dxy ) + (sin(fract(time*0.5f)*2.0f*3.14f)+1.0f)*0.5f );
    return vec4( plane_col.yzx * dv, 1.0f - u_Dissolve );
}
#endif

void main()
{
    vec3 rD = normalize( vec3( (start_pos.x + ax * v_texCoord.x),
                               (start_pos.y + ay * v_texCoord.y),
                               (start_pos.z                    ) ) );

    // was probably not a good idea to hardcode this

    float angle = atan( rD.x ) * (2.0f/16.0f);

    vec3 rO = rotate_xz( rot_xz,         22.0f );
    vec3 xz = rotate_xz( rot_xz + angle,  8.0f );

    rD = normalize( vec3( -xz.x, rD.y, -xz.z ) );

    vec3 light = rO;

    // inline intersect
    float b0 = dot( rO, rD );
    float d0 = b0 * b0 - dot( rO, rO ) + BOUNDING_RADIUS_2;
#if 0
    // redundant at correct zoom level
    if( d0 <= 0.0f ) {
        outColor  = vec4( 1.0f );
        return;
    }
#endif

	float t0 = -b0 - sqrt( d0 );

    rO += t0 * rD;

    float dist = intersectQJulia( rO, rD, mu_pos );

    if( dist >= EPSILON ) {
#ifdef XAVIER
		outColor  = getBG();
#else
		outColor  = vec4( 0.0f );
#endif
        return;
    }

    vec3 N = normEstimate( rO, mu_pos );

    outColor = vec4( Phong( light, rD, rO, N ), 1.0f - u_Dissolve );
}
